"""Game access token"""
from datetime import timedelta, datetime

from jose import jwt
from pydantic import BaseModel

from whist.core import SECRET_KEY, ALGORITHM
from whist.core.user.player import Player


class Token(BaseModel):
    """
    Handles jwt based token authentication for players during a game.
    """
    username: str

    @staticmethod
    def create(data: dict, expires_delta: timedelta = timedelta(minutes=15)) -> str:
        """
        Creates a token string.
        :param data: Payload of the token. Must contain 'sub:<username>'.
        :type data: dictionary
        :param expires_delta: The amount of time until the token expires. Default is 15 minutes.
        :type expires_delta: datetime.timedelta
        :return: the token string
        :rtype: string
        """
        to_encode = data.copy()
        expire = datetime.utcnow() + expires_delta
        to_encode.update({'exp': expire})
        return jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)

    @staticmethod
    async def get_user(database: dict, token: str) -> Player:
        """
        Asynchron method. Retrieves the user from a given database by token.
        :param database: Where to look for the user.
        :type database: dictionary
        :param token: From which the user will be extracted.
        :type token: string
        :return: A player instance if the user is found. Or raises a KeyError if the no username
        is in the token. Or raises ValueError if the user not in the database.
        :rtype: Player
        """
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        username: str = payload.get('sub')
        if username is None:
            raise KeyError('No username in token.')
        token = Token(username=username)
        player = Player.get_player(database, token.username)
        if player is None:
            raise ValueError(f'No user with username: {username} found.')
        return player



"""Errors in game phase"""
from whist.core.user.player import Player


class CardNotInHandError(Exception):
    """
    Raised when a player tries to play a card, that is not in their hand.
    """


class NotPlayersTurnError(Exception):
    """
    Raised when players tries to play although is not their turn.
    """

    def __init__(self, player: Player, turn_player: Player):
        super().__init__()
        self.message = f'Is not {player} turn, but {turn_player}.'


class NoTrumpSelectedError(Exception):
    """
    Raised when no trump is selected for hand.
    """


class TrickDoneError(Exception):
    """
    Raised when the trick is already done.
    """



"""DAO of session."""

from whist.core.error.table_error import TableFullError, TeamFullError, TableNotReadyError, \
    TableNotStartedError
from whist.core.game.rubber import Rubber
from whist.core.session.matcher import Matcher
from whist.core.session.session import Session
from whist.core.user.player import Player


class Table(Session):
    """
    The game logics instance of a room to play Whist.
    """
    min_player: int
    max_player: int
    team_size: int = 2
    started: bool = False
    rubbers: list[Rubber] = []

    # pylint: disable=too-few-public-methods
    class Config:
        """
        Configures the table class to allow private field. PrivateAttr cannot be used here as
        pylint does not detect the correct types in python 3.10.
        """
        underscore_attrs_are_private = True

    def __len__(self):
        """
        The amount of players joined.
        :return: # player
        :rtype: int
        """
        return len(self.users)

    @property
    def ready(self) -> bool:
        """
        Flag if the table is ready to start playing.
        :return Ready or not
        :rtype: boolean
        """
        return len(self.users) >= self.min_player and self.users.ready

    @property
    def current_rubber(self) -> Rubber:
        """
        Returns the current rubber
        :return: the latest rubber entry
        """
        if not self.started:
            raise TableNotStartedError()
        return self.rubbers[-1]

    def start(self, matcher: Matcher) -> None:
        """
        Starts the table, but will check if every player is ready first.
        """
        if not self.ready:
            raise TableNotReadyError()

        team_numbers = 2
        players_available_per_team = int(len(self.users) / team_numbers)
        teams = matcher.distribute(num_teams=team_numbers,
                                   team_size=min(players_available_per_team, self.team_size),
                                   users=self.users)
        rubber = Rubber(teams=teams)
        self.rubbers.append(rubber)
        self.started = True

    def join(self, player: Player) -> None:
        """
        If a seat is available a player joins the table.
        :param player: who wants to join the table
        :type player: Player
        :return: None or raised an error if the table is already full.
        :rtype: None
        """
        if len(self.users) < self.max_player:
            self.users.append(player)
        else:
            raise TableFullError(f'Table with name: {self.name} is already full.')

    def leave(self, player: Player) -> None:
        """
        Remove a player from table.
        :param player: The player to remove.
        :type player: Player
        :return: None
        :rtype: None
        """
        self.users.remove(player)

    def join_team(self, player: Player, team: int) -> None:
        """
        Player joins a team.
        :param player: to join a team
        :type player: Player
        :param team: id of the new team
        :type team: int
        :return: None if successful or raises Error if team is full
        :rtype: None
        """
        if not self.users.is_joined(player):
            self.join(player)
        team_size = self.users.team_size(team)
        if team_size >= self.team_size:
            raise TeamFullError(f'Team with id: {team} is already full.')
        self.users.change_team(player, team)

    def player_ready(self, player) -> None:
        """
        Player says they is ready.
        :param player: player who is ready
        :type player: Player
        :return: None
        :rtype: None
        """
        self.users.player_ready(player)

    def player_unready(self, player) -> None:
        """
        Player says they is not ready.
        :param player: player who is not ready
        :type player: Player
        :return: None
        :rtype: None
        """
        self.users.player_unready(player)



"""
Match making tool.
"""
import abc
import random

from whist.core.error.matcher_error import NotEnoughPlayersError
from whist.core.scoring.team import Team
from whist.core.session.userlist import UserList


# pylint: disable=too-few-public-methods
class Matcher(abc.ABC):
    """
    Abstrakt class for player to teams matching.
    """

    @staticmethod
    @abc.abstractmethod
    def distribute(num_teams: int, team_size: int, users: UserList) -> list[Team]:
        """
        Distributes cards according to subclass implementation.
        :param num_teams: the amount of teams
        :param team_size: how many players per team
        :param users: the user list at that table
        :return: the list of teams with players distributed to them
        """
        raise NotImplementedError


class RoundRobinMatcher(Matcher):
    """
    Distributes the players in the order of the user list.
    """

    @staticmethod
    def distribute(num_teams: int, team_size: int, users: UserList) -> list[Team]:
        """
        Distributes one player to each team each round in order of the user list. Repeats until
        the user list is empty.
        :param num_teams: the amount of teams
        :param team_size: how many players per team
        :param users: the user list at that table
        :return: the teams in round robin distribution
        """
        if team_size <= 0:
            raise ValueError('The team size must be positive.')
        if num_teams <= 0:
            raise ValueError('There must be at least one team.')
        players = users.players
        if num_teams * team_size > len(players):
            raise NotEnoughPlayersError(f'{num_teams * team_size} of players are need, but only '
                                        f'{len(players)} have joined.')
        for _ in range(0, team_size):
            for team_id in range(0, num_teams):
                users.change_team(players.pop(0), team_id)

        return users.teams


class RandomMatcher(Matcher):
    """
    Distributes the players randomly to teams.
    """

    @staticmethod
    def distribute(num_teams: int, team_size: int, users: UserList) -> list[Team]:
        """
        For given parameter distributes the players to teams.
        :return: None
        :rtype: None
        """
        players = users.players
        teams: list = list(range(0, team_size)) * num_teams
        for player in players:
            team_id = random.choice(teams)
            users.change_team(player, team_id)
            teams.remove(team_id)
        return users.teams



"""One Game of whist"""
from typing import Optional

from pydantic import BaseModel

from whist.core.game.hand import Hand
from whist.core.game.play_order import PlayOrder
from whist.core.game.player_at_table import PlayerAtTable
from whist.core.scoring.score import Score
from whist.core.scoring.score_card import ScoreCard
from whist.core.scoring.team import Team
from whist.core.user.player import Player


class Game(BaseModel):
    """
    One Game of whist.
    """

    play_order: PlayOrder
    win_score: int = 3
    score_card: ScoreCard = ScoreCard()
    current_hand: Optional[Hand] = None

    # pylint: disable=too-few-public-methods
    class Config:
        """
        Enables to have non models as field types and sets the encoder for play order.
        """
        arbitrary_types_allowed = True

    def next_hand(self) -> Hand:
        """
        Checks if the current hand is done and if so will return the next hand. If not it will
        return the current hand.
        :rtype: Hand
        """
        if self.current_hand is None:
            self.current_hand = Hand()
        elif self.current_hand.done():
            score = self._calc_score()
            self.score_card.add_score(score)
            self._next_play_order()
            self.current_hand = Hand()
        return self.current_hand

    def _calc_score(self):
        tricks_won = [0, 0]
        for trick in self.current_hand.tricks:
            winner = trick.winner
            tricks_won[winner.team] += 1
        tricks_won = [max(0, trick_score - 6) for trick_score in tricks_won]
        players_by_team = [[], []]
        for player in self.play_order:
            players_by_team[player.team].append(player.player)
        teams = [Team(players=players) for players in players_by_team]
        score = Score(teams=teams, scores=tricks_won)
        return score

    @property
    def current_trick(self):
        """
        Returns the current trick of the current hand.
        """
        try:
            return self.next_hand().current_trick
        except IndexError:
            return self.next_hand().deal(self.play_order)

    @property
    def done(self):
        """
        Check if game is done.
        :return: True if done else false
        :rtype: bool
        """
        return self.win_score <= self.score_card.max

    def get_player(self, player: Player) -> PlayerAtTable:
        """
        Retrieves the PlayerAtTable for the player given.
        :param player: who needs it's counterpart at the table
        :return: the player at table
        """
        return self.play_order.get_player(player)

    def _next_play_order(self) -> None:
        """
        Creates the next order of player for next hand.
        """
        self.play_order = self.play_order.next_order()



"""Hand of whist"""
from typing import Optional

from pydantic import BaseModel

from whist.core.cards.card import Card, Suit
from whist.core.cards.card_container import UnorderedCardContainer
from whist.core.error.hand_error import HandAlreadyDealtError
from whist.core.game.play_order import PlayOrder
from whist.core.game.player_at_table import PlayerAtTable
from whist.core.game.trick import Trick
from whist.core.game.warnings import TrickNotDoneWarning
from whist.core.util import enforce_str_on_dict


class Hand(BaseModel):
    """
    Hand of whist.
    """
    tricks: list[Trick] = []
    trump: Optional[Suit] = None

    def done(self) -> bool:
        """
        Check if the hand is done.
        :return: True if the hand is done, else False
        :rtype: bool
        """
        return len(self.tricks) == 13 and self.tricks[-1]

    @property
    def current_trick(self):
        """
        Returns the current trick.
        """
        return self.tricks[-1]

    def deal(self, play_order: PlayOrder) -> Trick:
        """
        Deals the hand and starts the first trick.
        :return: the first trick
        :rtype: Trick
        """
        if len(self.tricks) != 0:
            raise HandAlreadyDealtError()

        deck = UnorderedCardContainer.full()
        card: Optional[Card] = None
        while deck:
            player = play_order.get_next_player()
            card = deck.pop_random()
            player.hand.add(card)
        self.trump = card.suit

        first_trick = Trick(play_order=list(play_order), trump=self.trump)
        self.tricks.append(first_trick)
        return first_trick

    def next_trick(self, play_order: PlayOrder) -> Trick:
        """
        Starts the next trick.
        :return: the next trick
        :rtype: Trick
        """
        if not self.tricks[-1].done:
            raise TrickNotDoneWarning()
        next_trick_order = self._winner_plays_first_card(play_order)
        next_trick = Trick(play_order=list(next_trick_order), trump=self.trump)
        self.tricks.append(next_trick)
        return next_trick

    def dict(self, *args, **kwargs):
        super_dict = super().dict(*args, **kwargs)
        return enforce_str_on_dict(super_dict, ['trump'])

    def _winner_plays_first_card(self, play_order: PlayOrder) -> PlayOrder:
        winner: PlayerAtTable = self.tricks[-1].winner
        return play_order.rotate(winner)



"""Rubber of whist"""
from pydantic import BaseModel

from whist.core.game.game import Game
from whist.core.game.play_order import PlayOrder
from whist.core.scoring.team import Team
from whist.core.session.matcher import RandomMatcher
from whist.core.session.userlist import UserList


class Rubber(BaseModel):
    """
    Implementation of a rubber.
    """
    max_games: int = 3
    games: list[Game] = []
    teams: list[Team]

    # pylint: disable=too-few-public-methods
    class Config:
        """
        Configuration class for base model to allow fields which are not pydantic models.
        """
        arbitrary_types_allowed = True

    @property
    def games_played(self) -> int:
        """
        Amounts of games played already.
        :rtype: int
        """
        return len(self.games)

    @property
    def done(self) -> bool:
        """
        Checks if the rubber is done.
        :return: True if done else False
        :rtype: bool
        """
        return self.games_played == self.max_games

    def next_game(self) -> Game:
        """
        Creates a new game if the previous is done. Else returns the current game.
        :rtype: Game
        """
        if len(self.games) == 0 or self.games[-1].done:
            self.games.append(Game(play_order=PlayOrder.from_team_list(self.teams)))
        return self.games[-1]

    @classmethod
    def create_random(cls, users: UserList, num_teams: int, team_size: int) -> 'Rubber':
        """
        Creates a rubber with a random distribution of players in teams.
        :param users: all players and stati at the table
        :param num_teams: the amount of teams
        :param team_size: the size of each team
        :return: the rubber object
        """
        teams = RandomMatcher.distribute(num_teams, team_size, users)
        rubber = Rubber(teams=teams)
        return rubber



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

with open('requirements/requirements.txt') as f:
    requirements = f.readlines()

setup(
    name='whist-core',
    version='0.1.0',
    author='Whist Team',
    description='Game implementation of Whist.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Whist-Team/Whist-Core',
    project_urls={
        'Bug Tracker': 'https://github.com/Whist-Team/Whist-Core/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    keywords='game whist',
    packages=find_packages(exclude=('tests*',)),
    namespace_package=['whist'],
    python_requires='>=3.9',
    install_requires=requirements,
    extras_require={
        "testing": [
            "pytest==6.2.5",
            "pytest-cov==3.0.0",
            "pytest-asyncio==0.17.2"
        ]
    },
)
